#!/usr/bin/python

from nilmdb.utils.printf import *

import threading
import multiprocessing
import cStringIO
import sys
import os
import time
import uuid
import traceback

class LogReceiver(object):
    """Spawn a thread that listens to a pipe for log messages,
    and stores them locally."""
    def __init__(self, pipe):
        self.pipe = pipe
        self.log = cStringIO.StringIO()
        self.thread = threading.Thread(target = self.run)
        self.thread.start()

    def run(self):
        try:
            while True:
                data = self.pipe.recv_bytes()
                self.log.write(data)
        except EOFError:
            print "thread pipe died"
            self.pipe.close()
            print "thread closed"
            time.sleep(1)
            return

    def getvalue(self):
        return self.log.getvalue()

    def clear(self):
        self.log.truncate()
        self.log.seek(0)

class LogSender(object):
    """File-like object that writes output to a pipe"""
    def __init__(self, pipe):
        self.pipe = pipe

    def close(self):
        if self.pipe:
            self.pipe.close()
        self.pipe = None

    def write(self, data):
        if self.pipe:
            self.pipe.send_bytes(data)

    def flush(self, data):
        pass

    def isatty(self):
        return False

class Process(object):
    """Spawn and manage a running process"""
    def __init__(self, name, function, parameters):
        self.parameters = parameters
        self.start_time = None
        self.name = name
        pipes = multiprocessing.Pipe(duplex = False)
        self._log = LogReceiver(pipes[0])
        self._process = multiprocessing.Process(target = self._tramp,
                                                args = (function, pipes),
                                                name = name)
        self._process.daemon = True
        self._process.start()
        pipes[1].close()
        self.start_time = time.time()
        self.pid = str(uuid.uuid1(self._process.pid or 0))

    def _tramp(self, function, pipes):
        # Remap stdio in the child before calling function
        print "pid:", os.getpid()
        print "pipes:", pipes
        print "pipes[0].fileno():", pipes[0].fileno()
        print "pipes[1].fileno():", pipes[1].fileno()
        pipes[0].close()
        logfile = LogSender(pipes[1])

        saved_stdin = sys.stdin
        saved_stdout = sys.stdout
        saved_stderr = sys.stderr
        sys.stdin = open(os.devnull, 'r')
        sys.stdout = logfile
        sys.stderr = logfile

        exitcode = 0
        try:
            function(self.parameters)
            raise Exception("oh no")
        except:
            traceback.print_exc()
            exitcode = 1
        sys.stdin = saved_stdin
        sys.stdout = saved_stdout
        sys.stderr = saved_stderr
        logfile.close()
        sys.exit(exitcode)

    def terminate(self, timeout = 3):
        self._process.join(timeout)
        if self.alive:
            self._process.terminate()

    def clear_log(self):
        self._log.clear()

    @property
    def log(self):
        return self._log.getvalue()

    @property
    def alive(self):
        return self._process.is_alive()

    @property
    def error(self):
        if self._process.exitcode:
            return True
        return False

class ProcessManager(object):
    def __init__(self):
        self.processes = {}

    def __iter__(self):
        return iter(self.processes.keys())

    def __getitem__(self, key):
        return self.processes[key]

    def run(self, name, function, parameters):
        new = Process(name, function, parameters)
        self.processes[new.pid] = new
        return new.pid
